/*
 * Copyright 2011 Timothy Sparg
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.kilnhg.timsparg.juniorkilnclient;

import com.kilnhg.timsparg.juniorkilnclient.spi.KilnClientException;
import com.google.gson.Gson;
import com.kilnhg.timsparg.juniorkilndomain.KilnAPIErrorHolder;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Array;
import java.math.BigInteger;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.client.utils.URIUtils;
import org.apache.http.client.utils.URLEncodedUtils;
import org.apache.http.impl.client.BasicResponseHandler;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.message.BasicNameValuePair;

/**
 *
 * @author Tim Sparg
 */
public abstract class AbstractClient {

    protected final Gson gson = new Gson();
    protected final String kilnApiVersion;
    protected final URI baseURI;
    protected String authToken = null;
    protected final HttpClient httpclient = new DefaultHttpClient();

    protected AbstractClient(URI baseURI, String apiVersion) throws KilnClientException {
        kilnApiVersion = apiVersion;
        this.baseURI = baseURI;
    }

    protected String buildResource(String snippet) {
        if (baseURI.getPath() == null) {
            return "/API/" + kilnApiVersion + snippet;
        } else {
            if (baseURI.getPath().endsWith("/")) {
                return baseURI.getPath() + "API/" + kilnApiVersion + snippet;
            } else {
                return baseURI.getPath() + "/API/" + kilnApiVersion + snippet;
            }
        }
    }

    protected String toHex(String arg) {
        try {
            return String.format("%x", new BigInteger(arg.getBytes("UTF-8")));
        } catch (UnsupportedEncodingException ex) {
            throw new RuntimeException("Exception caused trying to Find UTF-8 charset", ex);
        }
    }

    protected void authenticate(String user, char[] pwd) throws KilnClientException {
        if (authToken != null) {
            return;
        }
        String result = kilnGet(KilnResource.AUTH_LOGIN.getSnippet(),
                new BasicNameValuePair("sUser", user),
                new BasicNameValuePair("sPassword", new String(pwd)));
        Arrays.fill(pwd, ' ');

        authToken = createObject(result, String.class);
    }

    protected String kilnGet(String snippet, BasicNameValuePair... params)
            throws KilnClientException {
        return callKiln(snippet, false, params);
    }

    protected String kilnPost(String snippet, BasicNameValuePair... params)
            throws KilnClientException {
        return callKiln(snippet, true, params);
    }

    protected String callKiln(String snippet, boolean isPost,
            BasicNameValuePair... params) throws KilnClientException {
        try {
            //independent copy so that I can add the token
            List<BasicNameValuePair> paramList = new ArrayList<BasicNameValuePair>();
            if (params != null) {
                Collections.addAll(paramList, params);
            }

            if (authToken != null) {
                paramList.add(new BasicNameValuePair("token", authToken));
            }
            URI uri = URIUtils.createURI(baseURI.getScheme(),
                    baseURI.getHost(), baseURI.getPort(), buildResource(snippet),
                    URLEncodedUtils.format(paramList, "UTF-8"), null);

            HttpRequestBase request;
            if (isPost) {
                request = new HttpPost(uri);
            } else {
                request = new HttpGet(uri);
            }

            return httpclient.execute(request, new BasicResponseHandler());

        } catch (URISyntaxException ex) {
            throw new KilnClientException("URISyntax Exception occured while "
                    + "trying to communicate with Kiln", ex);
        } catch (IOException ex) {
            throw new KilnClientException("IO Exception occured while trying "
                    + "to communicate with Kiln", ex);
        }
    }

    protected BasicNameValuePair[] createParameters(BasicNameValuePair... nameValuePairs) {
        List<BasicNameValuePair> params = new ArrayList<BasicNameValuePair>();
        for (BasicNameValuePair basicNameValuePair : nameValuePairs) {
            if (basicNameValuePair.getValue() != null && !basicNameValuePair.getValue().isEmpty()) {
                params.add(basicNameValuePair);
            }
        }

        return params.toArray(new BasicNameValuePair[params.size()]);
    }

    protected <T> T createObject(String jsonString, Class<T> clazz) {
        if (jsonString.startsWith("{\"errors\":")) {
            KilnAPIErrorHolder errorHolder = gson.fromJson(jsonString, KilnAPIErrorHolder.class);
            throw new KilnClientException(errorHolder);
        }
        return gson.fromJson(jsonString, clazz);
    }

    protected <T> List<T> createList(String jsonString, Class<T> clazz) {
        T[] array = (T[]) Array.newInstance(clazz, 0);
        array = (T[]) createObject(jsonString, array.getClass());

        return Arrays.asList(array);
    }
}
